﻿/*An asynchronous call will execute and immediately continue running your code instead of waiting for the call to finish. 
 *This is particularly useful for programs that control many outputs in a very short amount of time.
 *
 * A program for controlling multiple LEDs or multiple servos at a fast rate are good examples that benefit from async calls.
 * 
 * There are 3 common ways of using asynchronous calls
 * 
 * Method 1: Waiting for an Async Call with End'function name'
 *          a. call Begin'function name' to make async call and store the return value of this call in a variable 'varName' of type IAsyncResult
 *              eg. IAsyncResult result = servo.BeginSetTargetPosition(180, null, null);
 *          b. do some other work
 *          c. call End'function name' to wait for the async call to end (async call throws exception if it didn't make it to your device)
 *              eg. servo.EndSetTargetPosition(result);
 *          
 * Method 2: Polling for Asynchronous Call Completion
 *          a. call Begin'function name' to make async call and store the return value of this call in a variable 'varName' of type IAsyncResult
 *              eg. IAsyncResult result = servo.BeginSetTargetPosition(180, null, null);
 *          b. while ( ! 'varName'.IsComplete() )
 *                do some other work
 *                
 * Method 3: Executing a Callback Method When an Async Call Completes
 *          a. call Begin'function name' to make async call
 *              eg. servo.BeginSetTargetPosition(180, CallBackFunctionDelegate, infoObject);
 *                  Note.the last argument is any object that passes information that is later to be used by the callback function.
 *                       We recommend you use this argument similar to the way used in the Async C# example
 *          b. when the async call is complete, the CallBackFunction will be run
 *          c. Inside your CallBackFunction, call End'function name' if you want to check if the async call made it to your device or not.
 *              eg. servo.EndSetTargetPosition(result);
 *                  Note. The End'function name' (eg. EndSetTargetPosition) has to be called on the same object(eg. servo) that made the Begin'function name' (eg. BeginSetTargetPosition).
 *                        Async C# example shows one way of doing this.
 *                        
 * This example uses Method 3
 */

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using Phidget22;
using System.Diagnostics;

namespace RCServo_async_Example
{
    public partial class Form1 : Form
    {
        const int NUMBER_OF_SERVOS = 16;
        const int SERIAL_NUMBER = -1;
        const int MIN_PULSEWIDTH = 800000;
        const int MAX_PULSEWIDTH = 2000000;
        const double POSITION_AT_MIN_PULSEWIDTH = 0;
        const double POSITION_AT_MAX_PULSEWIDTH = 180;

        const int PORT = 0;

        RCServo[] m_servos;
        Stopwatch m_stopWatch;

        public Form1()
        {
            InitializeComponent();
            Init();
        }

        private void Init()
        {
            m_servos = new RCServo[NUMBER_OF_SERVOS];
            m_stopWatch = new Stopwatch();

            for (int i = 0; i < NUMBER_OF_SERVOS; i++)
            {
                m_servos[i] = new RCServo();
                m_servos[i].Attach += m_servo_Attach;
                m_servos[i].Detach += m_servo_Detach;
                m_servos[i].TargetPositionReached += Form1_TargetPositionReachedUpdate;
                m_servos[i].DeviceSerialNumber = SERIAL_NUMBER;
                m_servos[i].HubPort = PORT;
                m_servos[i].Channel = i;
                m_servos[i].Open();
            }
        }

        void Form1_TargetPositionReachedUpdate(object sender, Phidget22.Events.RCServoTargetPositionReachedEventArgs e)
        {
            if (((RCServo)sender).Channel == 0)
            {
                m_stopWatch.Stop();
                timeElapsedStatusLabel.Text = "| Time Elapsed : " + (m_stopWatch.ElapsedMilliseconds / 1000.0).ToString();
                m_stopWatch.Reset();
            }
        }

        void m_servo_Detach(object sender, Phidget22.Events.DetachEventArgs e)
        {
            statusLabel.Text = "Detached";
        }

        void m_servo_Attach(object sender, Phidget22.Events.AttachEventArgs e)
        {
            RCServo servo = (RCServo)sender;

            statusLabel.Text = "Attached";

            servo.Voltage = RCServoVoltage.Volts_5_0;
            servo.MinPulseWidth = MIN_PULSEWIDTH;
            servo.MaxPulseWidth = MAX_PULSEWIDTH;
            servo.MinPosition = POSITION_AT_MIN_PULSEWIDTH;
            servo.MaxPosition = POSITION_AT_MAX_PULSEWIDTH;
            servo.VelocityLimit = servo.MaxVelocityLimit;
            servo.Acceleration = servo.MaxAcceleration;
            servo.TargetPosition = 0;
            servo.Engaged = true;
        }

        private void setTargetPositionButton_Click(object sender, EventArgs e)
        {
            try
            {
                double targetPosition = (double)numericUpDown1.Value;

                AsyncCallback asyncCallBackDelegate = new AsyncCallback(AsyncCallBackForSetTargetPosition);

                for (int i = 0; i < NUMBER_OF_SERVOS; i++)
                {
                    CallBackInfo callBackInfo = new CallBackInfo(m_servos[i], targetPosition);
                    m_servos[i].BeginSetTargetPosition(targetPosition, asyncCallBackDelegate, callBackInfo);
                    m_stopWatch.Start();
                }
            }
            catch (Exception exp)
            {
                MessageBox.Show(exp.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void AsyncCallBackForSetTargetPosition(IAsyncResult result)
        {
            try
            {
                CallBackInfo callBackInfo = (CallBackInfo)result.AsyncState;
                callBackInfo.Owner.EndSetTargetPosition(result);
            }
            catch (Exception exp)
            {
                MessageBox.Show(exp.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void syncButton_Click(object sender, EventArgs e)
        {
            try
            {
                double targetPosition = (double)numericUpDown1.Value;

                for (int i = 0; i < NUMBER_OF_SERVOS; i++)
                {
                    m_servos[i].TargetPosition = targetPosition;
                }
            }
            catch (Exception exp)
            {
                MessageBox.Show(exp.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            for (int i = 0; i < NUMBER_OF_SERVOS; i++)
            {
                m_servos[i].Attach -= m_servo_Attach;
                m_servos[i].Detach -= m_servo_Detach;
                m_servos[i].TargetPositionReached -= Form1_TargetPositionReachedUpdate;
            }
        }
    }

    /// <summary>
    /// Class created to pass information to the callback
    /// </summary>
    public class CallBackInfo
    {
        public RCServo Owner { get; private set; }
        public double IntendedTargetPosition { get; private set; }

        public CallBackInfo(RCServo servo, double intendedTargetPosition)
        {
            Owner = servo;
            IntendedTargetPosition = intendedTargetPosition;
        }
    }
}
